#include "LRUList.h"

#include <string>
#include <strstream>
#include <time.h>

#define DEBUG_TRACE

#ifdef DEBUG_TRACE
#include <iostream>
#define DEBUG_ONLY(a) a
#else
#define DEBUG_ONLY(a) 
#endif


LRUList::LRUList(
   const size_t maxServants)
   :  m_numServants(0),
      m_maxServants(maxServants)
{
}

LRUList::~LRUList()
{
}
      
void LRUList::Evict()
{
   if (m_lruList.empty())
   {
      throw 0;
   }
   
   const size_t index = m_lruList.front();

   m_servants[index]->_remove_ref();

   m_lruList.pop_front();
}

PortableServer::ObjectId_var LRUList::Add(PortableServer::RefCountServantBase *pNewObject)
{
   int index;

   if (!m_freeList.empty())
   {
      index = m_freeList.front();
      m_freeList.pop_front();
   }
   else if (m_numServants < m_maxServants)
   {
      m_servants.push_back(0);

      index = m_servants.size() - 1;

      m_numServants++;
   }
   else
   {
      // what to do when we run out of space?

      Evict();       // force eviction of least recently used servant.

      if (m_freeList.empty())
      {
         throw 0;
      }
   
      index = m_freeList.front();
      m_freeList.pop_front();
   }

   m_servants[index] = pNewObject;

   PortableServer::ObjectId_var oid = GenerateOid(index);

   UseServant(index);

   DisplayList();

   return oid;
}

void LRUList::Remove(const PortableServer::ObjectId &oid)
{
   const size_t index = OidToIndex(oid);

   m_servants[index] = 0;

   m_lruList.remove(index);

   m_freeList.push_back(index);

//   DisplayList();
}

PortableServer::RefCountServantBase *LRUList::operator[] (const PortableServer::ObjectId &oid) 
{
   const size_t index = OidToIndex(oid);

   return UseServant(index);
}

PortableServer::RefCountServantBase *LRUList::UseServant(const size_t index)
{
   if (m_lruList.empty() || m_lruList.back() != index)
   {
      m_lruList.remove(index);
      m_lruList.push_back(index);
   }
     
   return m_servants[index];
}

void LRUList::DisplayList() const
{
   std::cout << "----- List is: -----" << std::endl;
   std::cout << "Num Nodes = " << m_numServants << std::endl;
   std::cout << "Max Nodes = " << m_maxServants << std::endl;

   std::list<size_t>::const_iterator it;

   for (it = m_lruList.begin(); it != m_lruList.end(); ++it)
   {
      const size_t index = *it;

      std::cout << index << " - " <<  "[" << m_servants[index];
      std::cout << "] oid = [" << index << "] " << std::endl;
   }
   
   std::cout << "---- Free List: ----" << std::endl;

   for (it = m_freeList.begin(); it != m_freeList.end(); ++it)
   {
      std::cout << *it << std::endl;
   }

   std::cout << "--------------------" << std::endl;

}

PortableServer::ObjectId_var LRUList::GenerateOid(
   const size_t index) const
{
   std::strstream stream;

   stream << (index + 1) << std::ends;

   std::string result = stream.str();

   stream.rdbuf()->freeze(false);

   return PortableServer::string_to_ObjectId(result.c_str());
}

size_t LRUList::OidToIndex(const PortableServer::ObjectId &oid) const
{
	CORBA::String_var oid_string = PortableServer::ObjectId_to_string(oid);

   char *pOidStr = oid_string;

   size_t index = atol(pOidStr);

   if (index > m_servants.size() || index == 0)
   {
      throw CORBA::OBJECT_NOT_EXIST();
   }

   index = index - 1;

   return index;
}
