#ifndef __TEVICTOR_BASE__
#define __TEVICTOR_BASE__

#include <string>
#include <strstream>
#include <vector>
#include <list>

#include "LRUList.h"

//#define DEBUG_TRACE

#ifdef DEBUG_TRACE
#include <iostream>
#define DEBUG_ONLY(a) a
#else
#define DEBUG_ONLY(a) 
#endif

#ifndef __CORBA_H_EXTERNAL_GUARD__
#include <omniORB3/CORBA.h>
#endif

#include <omnithread.h>

#pragma warning(disable : 4355)  // 'this' : used in base member initializer list
#pragma warning(disable : 4786)  // identifier was truncated to '255' characters in the debug information

template<class T, class T_ptr>
class TEvictorBase;

template<class POA_Class, class T, class T_ptr> 
class TEvictor : 
   public virtual POA_Class,
   public TEvictorBase<T, T_ptr>
{
   public:

      typedef TEvictor<POA_Class, T, T_ptr> EvictorBase;

      // Our main purpose in life is to remove the ambiguity between our POA class's 
      // _this() and the ref counted base _this()...

      T_ptr _this() 
      {
         return TEvictorBase<T, T_ptr>::_this();
      }

      void destroy()
      {
         TEvictorBase<T, T_ptr>::destroy();
      }

   protected :

		TEvictor(   
         const std::string &objectName)
         : TEvictorBase<T, T_ptr>(objectName)
      {
      }
};

template<class T, class T_ptr>
class TEvictorBase : 
	public PortableServer::RefCountServantBase
{
	public:
		

		virtual ~TEvictorBase();

      T_ptr _this() 
      {
         return m_obj;
      }
      
		class ServantLocator : 
			public virtual POA_PortableServer::ServantLocator
		{
			public :

				ServantLocator()
               : m_lruList(10)
				{
				}

				virtual PortableServer::Servant preinvoke(
					const PortableServer::ObjectId &oid,
					PortableServer::POA_ptr poa,
					const char *operation,
					PortableServer::ServantLocator::Cookie &the_cookie)
				{
					return GetObjectByReference(oid);
				}

				virtual void postinvoke(
					const PortableServer::ObjectId &oid, 
					PortableServer::POA_ptr poa, 
					const char* operation, 
					PortableServer::ServantLocator::Cookie the_cookie, 
					PortableServer::Servant the_servant)
				{
					the_servant->_remove_ref();
				}

				PortableServer::ObjectId_var AddToLocator(
					PortableServer::RefCountServantBase *pNewObject)
				{
               omni_mutex_lock locker(m_mutex);

               DEBUG_ONLY(std::cout << ::GetCurrentThreadId() << ": AddToLocator(" << pNewObject << ") = " << index << endl);

               return m_lruList.Add(pNewObject);
				}

            void RemoveFromLocator(PortableServer::ObjectId_var oid)
            {
               DEBUG_ONLY(std::cout << ::GetCurrentThreadId() << ": RemoveFromLocator(" << oid << ")" << std::endl);

               omni_mutex_lock locker(m_mutex);

               m_lruList.Remove(oid);
            }

			private :

				PortableServer::RefCountServantBase *GetObjectByReference(
					const PortableServer::ObjectId &oid)
				{
               omni_mutex_lock locker(m_mutex);

               PortableServer::RefCountServantBase *pObj = m_lruList[oid];

               DEBUG_ONLY(std::cout << ::GetCurrentThreadId() << ": GetObjectByReference(" << oid << ") = " << pObj << std::endl);

					pObj->_add_ref();

					return pObj;
				}

            LRUList m_lruList;

            omni_mutex m_mutex;

				// No copies
				ServantLocator(const ServantLocator &rhs);
				ServantLocator &operator=(const ServantLocator &rhs);
		};

	protected :

      void destroy();
		
      TEvictorBase(   
         const std::string &objectName);

	private :

      PortableServer::ObjectId_var m_oid;
      T_ptr m_obj;
      
      PortableServer::POA_ptr GetPOA(const std::string &objectName);

		static PortableServer::POA_ptr s_poa;
		static ServantLocator s_servantLocator;
};

template<class T, class T_ptr>
PortableServer::POA_ptr TEvictorBase<T,T_ptr>::s_poa = 0;

template<class T, class T_ptr>
TEvictorBase<T,T_ptr>::ServantLocator TEvictorBase<T,T_ptr>::s_servantLocator;

template<class T, class T_ptr>
TEvictorBase<T,T_ptr>::TEvictorBase<T,T_ptr>(
   const std::string &objectName) 
   :  m_oid(s_servantLocator.AddToLocator(this)),
      m_obj(T::_narrow(GetPOA(objectName)->create_reference_with_id(m_oid, T::_PD_repoId)))
{ 
   DEBUG_ONLY(std::cout << "TEvictorBase() - " << this << endl); 
}

template<class T, class T_ptr>
TEvictorBase<T,T_ptr>::~TEvictorBase<T,T_ptr>() 
{ 
   DEBUG_ONLY(std::cout << "~TEvictorBase() - " << this << std::endl); 

   s_servantLocator.RemoveFromLocator(m_oid);
}

template<class T, class T_ptr>
void TEvictorBase<T,T_ptr>::destroy()
{
	DEBUG_ONLY(std::cout << "TEvictorBase(" << this << ")::destroy()" << std::endl);

	RefCountServantBase::_remove_ref();
}

template<class T, class T_ptr>
PortableServer::POA_ptr TEvictorBase<T,T_ptr>::GetPOA(const std::string &objectName)
{
	if (0 == s_poa)
	{
		int argc = 0;

      CORBA::ORB_var orb = CORBA::ORB_init(argc ,0,"");

		CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
		PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(obj);

		PortableServer::POAManager_var POAManager = rootPOA->the_POAManager();

		CORBA::PolicyList policyList;
		policyList.length(3);
		policyList[0] = rootPOA->create_request_processing_policy(PortableServer::USE_SERVANT_MANAGER);
		policyList[1] = rootPOA->create_servant_retention_policy(PortableServer::NON_RETAIN);
		policyList[2] = rootPOA->create_id_assignment_policy(PortableServer::USER_ID);

      std::string POAName = objectName + "::POA";

		s_poa = rootPOA->create_POA(POAName.c_str(), POAManager, policyList);

		// Activate a servant locator in the root poa.
    
		PortableServer::ObjectId_var id = rootPOA->activate_object(&s_servantLocator);
		PortableServer::ServantLocator_var slref = s_servantLocator._this();
		s_servantLocator._remove_ref();

		// Register the servant activator with our new poa.
		s_poa->set_servant_manager(slref);   
   }

	return s_poa;
}

#endif // __TEVICTOR_BASE__
